function [boxes, exvec, pyra] = detector(im, model, thresh, write, bbox, overlap)

INF = 1e10;
MAXB = 50000;
latent = false;

if ~exist('thresh', 'var')
  thresh = -INF;
end

if ~exist('write', 'var')
  write = false;
end

if ~exist('bbox', 'var') || isempty(bbox)
  bbox = [];
else
  latent = true;
end

if ~exist('overlap', 'var')
  overlap = 0.6;
end

% Compute the feature pyramid and prepare filter
pyra     = featpyramid(im,model);
interval = model.interval;
levels   = 1:length(pyra.feat);

% Cache various statistics derived from model
[components,filters,resp] = modelcomponents(model,pyra);
boxes     = zeros(MAXB,length(components{1})*4+2);
exvec(MAXB) = struct('id', [], 'blocks', []);
ex.id     = [0 0 0 0 0];
ex.blocks = [];
cnt       = 0;

rscores = cell(1, length(levels));

% Iterate over random permutation of scales and components,
for rlevel = levels,
  % Iterate through mixture components
  for c  = randperm(length(model.components)),
    parts    = components{c};
    numparts = length(parts);

    % Skip if there is no overlap of root filter with bbox
    if latent,
      skipflag = 0;
      for k = 1:numparts
        % because all mixtures for one part is the same size, we only need to do this once
        skipflag = 1;
        
        for fi = 1:length(parts(k).filterid)
          ovmask = testoverlap(parts(k).sizx(fi),parts(k).sizy(fi),pyra,rlevel,bbox{k},overlap);
          if (any(ovmask(:)))
            skipflag = 0;
            break;
          end
        end

        if skipflag == 1
          break;
        end
      end
      if skipflag == 1
        continue;
      end
    end

    % Local scores
    for k = 1:numparts,
      level = rlevel-parts(k).scale*interval;
      if isempty(resp{level})
        resp{level} = fconv(pyra.feat{level},filters,1,length(filters));
        
        padsize = model.maxsize-1;
        featsize = size(pyra.feat{level});
        alignsize = featsize(1:2)-model.maxsize+1;
        for i = 1:length(filters)
          offsetx1 = ceil(size(filters{i}, 2)/2)-1;
          offsety1 = ceil(size(filters{i}, 1)/2)-1;
          resp{level}{i} = circshift(resp{level}{i}, [offsety1, offsetx1]);
          resp{level}{i} = resp{level}{i}(padsize(1)+1:alignsize(1), padsize(2)+1:alignsize(2));
        end
      end
      
      f = parts(k).filterid;
      parts(k).score = zeros([size(resp{level}{1}), parts(k).ctnum]);
      for ti = 1:parts(k).ctnum
        fi = parts(k).ctmap(ti);
        parts(k).score(:,:,ti) = resp{level}{f(fi)};
      end
      
      parts(k).level = level;
	  
      if latent
        for fi = 1:length(f)
          ovmask = testoverlap(parts(k).sizx(fi),parts(k).sizy(fi),pyra,rlevel,bbox{k},overlap);
          i = f(fi);
          featsize = size(pyra.feat{level});
          alignsize = featsize(1:2)-model.maxsize+1;
          padsize = model.maxsize-1;
          offsetx1 = ceil(size(filters{i}, 2)/2)-1;
          offsety1 = ceil(size(filters{i}, 1)/2)-1;
          ovmask = circshift(ovmask, [offsety1, offsetx1]);
          ovmask = ovmask(padsize(1)+1:alignsize(1), padsize(2)+1:alignsize(2));
          tmpscore = parts(k).score(:,:,fi);
          tmpscore(~ovmask) = -INF;
          parts(k).score(:,:,fi) = tmpscore;
        end
      end
      
    end
    
    % Walk from leaves to root of tree, passing message to parent
    for k = numparts:-1:2,
      par = parts(k).parent;
      [msg,parts(k).Ix,parts(k).Iy,parts(k).Ik] = passmsg(parts(k),parts(par));
      parts(par).score = parts(par).score + msg;
    end

    % Add bias to root score
    parts(1).score = parts(1).score + parts(1).b;
    [rscore Ik]    = max(parts(1).score,[],3);
    rscores{level} = rscore;
    
    % Zero-out invalid regions in latent mode
    if latent,
      thresh = max(thresh,max(rscore(:)));
    end
    
    thresh = max(thresh, -INF/2);
    [Y,X] = find(rscore >= thresh);
    
    % Walk back down tree following pointers
    % (DEBUG) Assert extracted feature re-produces score
    for i = 1:length(X)
      x = X(i);
      y = Y(i);
      k = Ik(y,x);
      
      [box,ex,ptr] = backtrack( x , y , k, parts , pyra , ex , latent || write);
      
      cnt = cnt + 1;
      boxes(cnt,:) = [box level rscore(y,x)];
      exvec(cnt) = ex;
      
      if (cnt > MAXB)
        break;
      end
    end
  end
  
  if (cnt > MAXB)
    break;
  end
end

boxes = boxes(1:cnt,:);
exvec = exvec(1:cnt);


% ----------------------------------------------------------------------
% Helper functions for detection, feature extraction, and model updating
% ----------------------------------------------------------------------

% Cache various statistics from the model data structure for later use  
function [components,filters,resp] = modelcomponents(model,pyra)

  components = cell(length(model.components),1);
  for c = 1:length(model.components),
    for k = 1:length(model.components{c}),
      p = model.components{c}(k);
      [p.sizy,p.sizx,p.w,p.defI,p.starty,p.startx,p.step,p.level,p.Ix,p.Iy] = deal([]);
      [p.scale,p.level,p.Ix,p.Iy] = deal(0);
 
      % store the scale of each part relative to the component root
      par = p.parent;      
      assert(par < k);
      p.b = [model.bias(p.biasid).w];
      p.b = reshape(p.b,[1 size(p.biasid)]);
      p.biasI = [model.bias(p.biasid).i];
      p.biasI = reshape(p.biasI,size(p.biasid));
      
      % =============
      p.fws = [];
      p.fwsI = [];
      
      for f = 1:length(p.filterid)
        x = model.filters(p.filterid(f));
        [p.sizy(f) p.sizx(f) foo] = size(x.w);
        p.filterI(f) = x.i;
      end
      
      for f = 1:numel(p.defid)	  
        x = model.defs(p.defid(f));
        p.w(:,f)  = x.w';
        p.defI(f) = x.i;
        ax  = x.anchor(1);
        ay  = x.anchor(2);    
        ds  = x.anchor(3);
        p.scale = ds + components{c}(par).scale;
        % amount of (virtual) padding to hallucinate
        step     = 2^ds;
        virtpady = (step-1)*pyra.pady;
        virtpadx = (step-1)*pyra.padx;
        % starting points (simulates additional padding at finer scales)
        p.starty(f) = ay-virtpady;
        p.startx(f) = ax-virtpadx;      
        p.step   = step;
      end
      components{c}(k) = p;
    end
  end
  
  resp    = cell(length(pyra.feat),1);
  filters = cell(length(model.filters),1);
  for i = 1:length(filters),
    filters{i} = model.filters(i).w;
  end
  
% Given a 2D array of filter scores 'child',
% (1) Apply distance transform
% (2) Shift by anchor position of part wrt parent
% (3) Downsample if necessary
function [score,Ix,Iy,Ik] = passmsg(child,parent)

INF = 1e10;
INF2 = 2e10;
K = child.ctnum;
L = child.ptnum;

Ny  = size(parent.score,1);
Nx  = size(parent.score,2);

N  = Nx*Ny;
i0 = reshape(1:N,Ny,Nx);
[score,Ix,Iy,Ik] = deal(zeros(Ny,Nx,L));

for li = 1:L
  
  Ix0 = zeros([Ny Nx K]);
  Iy0 = zeros([Ny Nx K]);
  score0 = -INF*ones([Ny Nx K]);

  for ki = 1:K
    fltid = child.ctmap(ki);
    defid = L*(ki-1)+li;
    
    if child.vtmap(li,ki) == false
      score0(:,:,ki) = -INF2 *ones([Ny Nx]);
      continue;
    end
    
    [score0(:,:,ki),Ix0(:,:,ki),Iy0(:,:,ki)] = shiftdt(child.score(:,:,fltid), ...
      child.w(1,defid), child.w(2,defid), child.w(3,defid), child.w(4,defid), ...
      child.startx(defid),child.starty(defid),Nx,Ny,child.step);
    
  end
  
  b = child.b(1,li,:);
  score0 = bsxfun(@plus, score0, b);
  
  % At each parent location, for each parent mixture 1:L, compute best child mixture 1:K
  [score(:,:,li),I] = max(score0(:,:,:),[],3);
  
  i = i0 + N*(I-1);
  Ix(:,:,li)    = Ix0(i);
  Iy(:,:,li)    = Iy0(i);
  Ik(:,:,li)    = I;
  
end

% Compute a mask of filter reponse locations (for a filter of size sizy,sizx)
% that sufficiently overlap a ground-truth bounding box (bbox) 
% at a particular level in a feature pyramid
function ov = testoverlap(sizx,sizy,pyra,level,bbox,overlap)
  scale = pyra.scale(level);
  padx  = pyra.padx;
  pady  = pyra.pady;
  [dimy,dimx,~] = size(pyra.feat{level});
  
  bx1 = bbox(1);
  by1 = bbox(2);
  bx2 = bbox(3);
  by2 = bbox(4);
  
  % Index windows evaluated by filter (in image coordinates)
  x1 = ((1:dimx-sizx+1) - padx - 1)*scale + 1;
  y1 = ((1:dimy-sizy+1) - pady - 1)*scale + 1;
  x2 = x1 + sizx*scale - 1;
  y2 = y1 + sizy*scale - 1;
  
  % Compute intersection with bbox
  xx1 = max(x1,bx1);
  xx2 = min(x2,bx2);
  yy1 = max(y1,by1);
  yy2 = min(y2,by2);
  w   = xx2 - xx1 + 1;
  h   = yy2 - yy1 + 1;
  w(w<0) = 0;
  h(h<0) = 0;
  inter  = h'*w;
  
  % area of (possibly clipped) detection windows and original bbox
  area = (y2-y1+1)'*(x2-x1+1);
  box  = (by2-by1+1)*(bx2-bx1+1);
  
  % thresholded overlap
  ov   = inter ./ (area + box - inter) > overlap;

  

