function [idx, realK] = learnVisualCategory(pos, boxes, scale, K, kmean_iter, sbin)
warning off;

if (K == 1)
  idx = ones(1, length(pos));
  realK = 1;
  return;
end

neg = INRIA_data();
neg = neg(1:min(length(neg),100));

%% parameter setting
npos = length(pos);
ind_te = randperm(npos);
ind_tr = ind_te(1:floor(npos/2));
ind_te = ind_te(floor(npos/2)+1:end);

tr_id = 1;
te_id = 2;
splitset(1).ind = ind_tr;
splitset(1).pos = pos(ind_tr);
splitset(1).box = boxes(ind_tr, :);
splitset(1).npos = length(ind_tr);
splitset(1).idx = [];

splitset(2).ind = ind_te;
splitset(2).pos = pos(ind_te);
splitset(2).box = boxes(ind_te, :);
splitset(2).npos = length(ind_te);
splitset(2).idx = [];

fdim = 32;
maxiter = 5;
Nmax = 5000; % including pos samples and neg samples
nneg = length(neg);
nrnd = 20;

% set up sub models
modelnum = K;
flags = ones(1, modelnum);
submodel = cell(1, modelnum);
b = zeros(2, modelnum);


%% build bboxes for positive samples
boxsize = [boxes(:,4)-boxes(:,2), boxes(:,3)-boxes(:,1)]./[scale', scale'];
featsz = floor(([median(boxsize(1)), median(boxsize(2))])./sbin);
ptchsz = featsz*sbin;

%% learn visual categories using unsupervised method
% wrap training patches
ptchs = cell(1,npos);
feats = zeros([prod(featsz-2)*fdim, npos]);
for id = 1:npos
  im = imread(pos(id).im);
  ptch = crop_image(im, boxes(id,:), 1);
  ptchs{id} = imresize(ptch, ptchsz);
  feat = features(double(ptchs{id}), sbin);
  feat = feat(:,:,1:fdim);
  feats(:,id) = feat(:);
end
% collect negative samples
feats_neg = zeros([prod(featsz-2)*fdim, nneg*nrnd]);
ct = 1;
for n = 1:nneg
  im = imread(neg(n).im);
  [r,c,z] = size(im);
  negpos = ceil(bsxfun(@times, rand([nrnd,2]), ([r,c]-ptchsz)));
  for ri = 1:nrnd
    feat = features(double(crop_image(im, [negpos(ri,[2,1]), negpos(ri,[2,1])+ptchsz-1], 1)), sbin);
    feat = feat(:,:,1:fdim);
    feats_neg(:, ct) = feat(:);
    ct = ct + 1;
  end
end

feats_tr = feats(:, splitset(tr_id).ind);
splitset(tr_id).idx = kmeans_wrap(feats_tr', K, kmean_iter);
modelnum = K;
  
%% do cross-validation
for cvi = 1 : maxiter  % max cross-validation iterations  
  idx_tr = splitset(tr_id).idx;
  
  for fcnt = 1:modelnum
    if ~flags(fcnt), continue; end
    
    feats_tr = feats(:, splitset(tr_id).ind);
    
    cind = find(idx_tr==fcnt);
    clen = length(cind);
    
    if (clen < 5) % skip clusters with less than 5 positive
      flags(fcnt) = 0;
      continue;
    end
    
    % add positive and negative samples
    posfeat = feats_tr(:,cind);
    data_vec = [posfeat, feats_neg]';
    label_vec = [ones(clen, 1); -ones(nneg*nrnd, 1)];
    submodel{fcnt} = trainlibsvm(label_vec, sparse(data_vec), ' -q ');
    
    % normalize score
    tscore = sum(bsxfun(@times,data_vec,submodel{fcnt}.w), 2);
    ylabel = [ones(size(posfeat,2), 1); zeros(size(feats_neg,2), 1)];
    [b(:, fcnt)] = glmfit(tscore,ylabel,'binomial','link','logit');
    
  end
  
  scores = repmat(-Inf, [splitset(te_id).npos, modelnum]);
  for fcnt = 1:modelnum
    if ~flags(fcnt), continue; end
    % submodels evaluation, update labels according to scores
    tscore = sum(bsxfun(@times,feats(:,splitset(te_id).ind)',submodel{fcnt}.w), 2);
    scores(:, fcnt) = glmval(b(:,fcnt), tscore, 'logit');
  end
  
  idx_tmp = zeros(splitset(te_id).npos, 1);
  idx_rest = [1:splitset(te_id).npos];
  while any(idx_tmp(:)==0),
    [tscore, idx_tmp_0] = partition_labels(scores(idx_rest, :));
    idx_tmp(idx_rest) = idx_tmp_0;
    idx_rest = find(idx_tmp == 0);
  end

  % converge test
  if ~isempty(splitset(te_id).idx)
    relabel = numel(find(splitset(te_id).idx ~= idx_tmp));
    if (relabel < 2)
      splitset(te_id).idx = idx_tmp;
      break;
    end
  end
  
  splitset(te_id).idx = idx_tmp;
  
  %swap
  tr_id = te_id;
  te_id = 3-tr_id;
end

scores = repmat(-Inf, [npos, modelnum]);
for fcnt = 1:modelnum
  if ~flags(fcnt), continue; end
  % submodels evaluation, update labels according to scores
  tscore = sum(bsxfun(@times,feats(:,:)',submodel{fcnt}.w), 2);
  scores(:, fcnt) = glmval(b(:,fcnt), tscore, 'logit');
end

idx_tmp = zeros(npos, 1);
idx_rest = [1:npos];
while any(idx_tmp(:)==0),
  [tscore, idx_tmp_0] = partition_labels(scores(idx_rest, :));
  idx_tmp(idx_rest) = idx_tmp_0;
  idx_rest = find(idx_tmp == 0);
end

% prepare for output
labels = unique(idx_tmp);
realK = length(labels); % update cluster number
idx = zeros(1, length(pos));
for i = 1:realK
  idx(idx_tmp==labels(i)) = i;
end


function [tscore, idx] = partition_labels(scores)
% partition N samples into M labels, according to test scores of M
% classifiers
% scores : NxM matrix, test scores of M classifiers

[N, M] = size(scores);
idx = zeros(N, 1);
tscore = zeros(N, 1);

%% parameters for partition
% classifier threshhold
thresh = min(0.5, max(scores(:))*0.5);
eta = repmat(thresh,1,M);
% top select number
topn = 20;

sflag = false(N, 1);
sele = [];
for i = 1:M
  sids = find(scores(:, i) >= eta(i));
  sids = setdiff(sids, sele);  % remove assigned labels
  slen = length(sids);
  tops = min(slen, topn);
  
  if (tops < slen)
    sids = sids(randperm(slen));
    sids = sids(1:tops);
  end
  
  idx(sids) = i;
  tscore(sids) = scores(sids, i);
  
  sflag(sids) = true;
  sele = union(sele, sids);
end





