function model = train(name, model, pos, neg, warp, iter, C, wpos, overlap) 
% model = train(name, model, pos, neg, warp, iter, C, wpos, overlap)
%               1,    2,     3,   4,   5,    6,    7, 8,    9
% Train a structured SVM with latent assignement of positive variables
% pos  = list of positive images with part annotations
% neg  = list of negative images
% warp = 1 uses warped positives
% warp = 0 uses latent positives
% iter is the number of training iterations
%   C  = scale factor for slack loss
% wpos = amount to weight errors on positives
% overlap = minimum overlap in latent positive search

global opts

if nargin < 6
  iter = 1;
end
if nargin < 7
  C = 0.002;
end
if nargin < 8
  wpos = 2;
end
if nargin < 9
  overlap = 0.6;
end

% set the size of the training data cache (in GB)
maxsize = opts.maxmem; 
fprintf('Using %.1f GB\n',maxsize);


% Vectorize the model
len  = sparselen(model);
nmax = round(maxsize*.25e9/len);

% Define global QP problem
clear global qp;
global qp;
% qp.x(:,i) = examples
% qp.i(:,i) = id
% qp.b(:,i) = bias of linear constraint
% qp.d(i)   = ||qp.x(:,i)||^2
% qp.a(i)   = ith dual variable
qp.x   = zeros(len,nmax,'single');
qp.i   = zeros(5,nmax,'int32');
qp.b   = ones(nmax,1,'single');
qp.d   = zeros(nmax,1,'double');
qp.a   = zeros(nmax,1,'double');
qp.sv  = logical(zeros(1,nmax));  
qp.n   = 0;
qp.obj = [];

[qp.w,qp.wreg,qp.w0,qp.noneg] = model2vec(model);
qp.Cpos = C*wpos;
qp.Cneg = C;
qp.w    = (qp.w - qp.w0).*qp.wreg;

for t = 1:iter,
  
  if ~isfield(model, 'thresh')
    model.thresh = 0;
  else
    model.thresh = model.thresh + 0.1;
  end
  
  fprintf('\niter: %d/%d\n', t, iter);
  qp.n = 0;
  if warp > 0
    numpositives = poswarp(name, t, model, pos);
  else
    numpositives = poslatent(name, t, model, pos, overlap);
  end
  
  for i = 1:length(numpositives)
    fprintf('component %d got %d positives\n', i, numpositives(i));
  end
  assert(qp.n <= nmax);
  
  % Fix positive examples as permenant support vectors
  % Initialize QP soln to a valid weight vector
  % Update QP with coordinate descent
  qp.svfix = 1:qp.n;
  qp.sv(qp.svfix) = 1;
  qp_prune();
  qp_opt(1);
  model = vec2model(qp_w,model);
  model.interval = 2;
  loss = 0;

  % grab negative examples from negative images
  for i = 1:length(neg),
    fprintf('\n Image(%d/%d)',i,length(neg));
    im  = imread(neg(i).im);
    [box,model,loss] = detect_warp(im, model, -1, [], 0, i, -1,loss);

    fprintf(' #cache+%d=%d/%d, #sv=%d, LB=%.4f, loss=%.4f',size(box,1),qp.n,nmax,sum(qp.sv),qp.obj,loss);
    % Stop if cache is full
    if sum(qp.sv) == nmax,
      break;
    end
  end

  % One final pass of optimization
  qp_opt();
  model = vec2model(qp_w,model);

  fprintf('\nDONE iter: %d/%d #sv=%d/%d, LB=%.4f\n',t,iter,sum(qp.sv),nmax,qp.obj);

  % Compute minimum score on positive example (with raw, unscaled features)
  r = sort(qp_scorepos);
  model.thresh   = r(ceil(length(r)*.05));
  model.interval = 5;
  model.obj = qp.obj;
end

fprintf('qp.x size = [%d %d]\n',size(qp.x));
clear global qp;

% get positive examples by warping positive bounding boxes
% we create virtual examples by flipping each image left to right
function numpositives = poswarp(name, t, model, pos)
  numpos = length(pos);
  
  f   = model.components{1}(1).filterid;
  siz = size(model.filters(f).w);
  siz = siz(1:2);
  warped = warppos(name, siz, model.sbin, pos);

  for i = 1:numpos
    fprintf('%s: iter %d: warped positive: %d/%d\n', name, t, i, numpos);
    % get example
    im = warped{i};
    feat = features(im, model.sbin);
    qp_poswrite(feat,i,model);
  end
  global qp;
  numpositives = qp.n;
  
function qp_poswrite(feat,id,model)
  ex.id     = [1 id 0 0 0]';
  ex.blocks = [];
  ex.blocks(end+1).i = model.bias.i;
  ex.blocks(end).x   = 1;
  ex.blocks(end+1).i = model.filters.i;
  ex.blocks(end).x   = feat;
  
  qp_write(ex, 0);
  

% get positive examples using latent detections
% we create virtual examples by flipping each image left to right
function numpositives = poslatent(name, t, model, pos, overlap)
  
  numpos = length(pos);
  model.interval = 2;
  numpositives = zeros(length(model.components), 1);
  
  % we add the below lines 
  numparts = length(model.components{1});
  
  for i = 1:numpos
    fprintf('%s: iter %d: latent positive: %d/%d\n', name, t, i, numpos);
    % skip small examples
    %skipflag = 0;
    
    spos = pos(i);
    
    for j = 1:length(spos)
      
      bbox = cell(1,numparts);
      for p = 1:numparts
        bbox{p} = [spos(j).x1(p) spos(j).y1(p) spos(j).x2(p) spos(j).y2(p)];
      end
      
      % get example
      if (ischar(spos(j).im))
        im = imread(spos(j).im);
      else
        im = spos(j).im;
      end
      [im, bbox] = croppos(im, bbox);
      box = detect_warp(im, model, model.thresh, bbox, overlap, i, 1);
      if ~isempty(box),
        fprintf(' (comp=%d,sc=%.3f)\n',box(1,end-1),box(1,end));
        numpositives = numpositives+1;
      end
      
    end
    
  end

% Compute score (weights*x) on positives examples (see qp_write.m)
% Standardized QP stores w*x' where w = (weights-w0)*r, x' = c_i*(x/r)
% (w/r + w0)*(x'*r/c_i) = (v + w0*r)*x'/ C
function scores = qp_scorepos
  global qp;
  y = qp.i(1,1:qp.n);
  I = find(y == 1);
  w = qp.w + qp.w0.*qp.wreg;
  scores = score(w,qp.x,I) / qp.Cpos;


% [newim, newbox] = croppos(im, box)
% Crop positive example to speed up latent search.
function [im, box] = croppos(im, box)

  if iscell(box) % many parts are stored in box
    P = length(box);
    x1 = zeros(1,P);
    y1 = zeros(1,P);
    x2 = zeros(1,P);
    y2 = zeros(1,P);
    for p = 1:P
      x1(p) = box{p}(1);
      y1(p) = box{p}(2);
      x2(p) = box{p}(3);
      y2(p) = box{p}(4);
    end
    x1 = min(x1); y1 = min(y1); x2 = max(x2); y2 = max(y2);
    pad = 0.5*((x2-x1+1)+(y2-y1+1));
    x1 = max(1, round(x1-pad));
    y1 = max(1, round(y1-pad));
    x2 = min(size(im,2), round(x2+pad));
    y2 = min(size(im,1), round(y2+pad));
    
    im = im(y1:y2, x1:x2, :);
    for p = 1:P
      box{p}([1 3]) = box{p}([1 3]) - x1 + 1;
      box{p}([2 4]) = box{p}([2 4]) - y1 + 1;
    end
    
  else % only one template is stored in box
       % crop image around bounding box
    pad = 0.5*((box(3)-box(1)+1)+(box(4)-box(2)+1));
    x1 = max(1, round(box(1) - pad));
    y1 = max(1, round(box(2) - pad));
    x2 = min(size(im, 2), round(box(3) + pad));
    y2 = min(size(im, 1), round(box(4) + pad));
    
    im = im(y1:y2, x1:x2, :);
    box([1 3]) = box([1 3]) - x1 + 1;
    box([2 4]) = box([2 4]) - y1 + 1;
  end
  
